/*
 * ForceLTSView.java
 *
 * Created on 30 July 2007, 00:07
 */

package assignment;

import java.util.*;
import java.lang.*;
import javax.swing.*;
import java.awt.*;
import java.awt.geom.Point2D;
import javax.swing.Timer;
import java.awt.event.*;

/**
 *
 * @author David
 */
public class ForceLTSView extends LTSView {
    /** This represents a node in the graph. It holds position, velocity and
     * other such stuff.
     */
    protected class graphNode {
        public graphVector velocity;
        public graphVector position;
        //public graphVector net_force;
        public LTSNode theNode;
        public ArrayList<graphNode> springs;
        public ArrayList<String> spring_names;
        public boolean noDraw;
        public graphNode() {
            velocity = new graphVector(0,0);
            position = new graphVector(0,0);
            //net_force = new graphVector(0,0);
            springs = new ArrayList<graphNode>();
            spring_names = new ArrayList<String>();
            noDraw = false;
        }
        public graphNode(LTSNode n) {
            velocity = new graphVector(0,0);
            position = new graphVector(0,0);
            //net_force = new graphVector(0,0);
            springs = new ArrayList<graphNode>();
            spring_names = new ArrayList<String>();
            theNode = n;
            noDraw = false;
        }
        public void randomizePosition() {
            position.x = rand.nextInt(getWidth());
            position.y = rand.nextInt(getHeight());
        }
    }
    
    protected Random rand;
    ArrayList<graphNode> theNodes;
    HashMap<String, graphNode> nodeMap = new HashMap<String,graphNode>();
    
    protected double repelconst = 8000;
    protected double springconst = 1.5;
    protected double timestep = 0.2;
    protected double damping = 0.45;
    
    javax.swing.Timer timer;
    protected int timerSpeed = 25;
    
    protected double total_kinetic_energy = 0;
    
    /** Creates a new instance of ForceLTSView */
    public ForceLTSView() {
        rand = new Random();
        theNodes = new ArrayList<graphNode>();
    }
    
    public void renderLTS() {
        //System.out.println("====== Rendering LTS ======");
        // Here we must begin the rendering process.
        
        /* The many steps to force-based success:
         * 1.1. Create nodes
         * 1.2. Set initial node velocities to (0,0)
         * 1.3. Set initial node positions randomly. No 2 nodes may be in the
         *      exact same place.
         */
        
        if (theLTS == null) return;
        
        if (timer != null) {
            if (timer.isRunning()){
                timer.stop();
                System.out.println("==== Render stopped ====");
                return;
            }
        }
        
        if (theLTS.getNodes().size() == 0) {
            System.out.println("Nothing to render.");
        }
        else {
            
            ActionListener al = new ActionListener(  ) {
                     public void actionPerformed(ActionEvent event) {
                         //System.out.println("======================================> Wow! the timer fired!");
                         total_kinetic_energy = 0;
                         total_kinetic_energy = iterate();
                         repaint();
                         if (total_kinetic_energy < 50){
                             timer.stop();
                             System.out.println("===== Render Complete =====");
                         }
                     }
                };

            timer = new Timer(timerSpeed, al);
            timer.start(); 
            System.out.println("==== Render Running ====");
        }
    }
    
    public void setup() {
        if (theLTS.getNodes().size() == 0) {
            System.out.println("Can't setup: no nodes.");
        }
        else {
            clearNodes();
            createNodes();
        }
    }
    
    protected void createNodes() {
        System.out.println("Creating nodes...");
        Collection<LTSNode> nodes = theLTS.getNodes();
        
        // We need to stick the start node in the middle.
        LTSNode anchorNode = new LTSNode();
        anchorNode.setName("$ANCHOR_NODE$$");
        graphNode gnn = new graphNode(anchorNode);
        gnn.position.x = getWidth() / 2;
        gnn.position.y = getHeight() / 2;
        gnn.noDraw = true;
       
        for(LTSNode n:nodes) {
            graphNode gn = createNode(n);
            storeNode(gn);
        }
        
        for (graphNode n:theNodes) {
            gnn.springs.add(n);
            n.springs.add(gnn);
        }
       
        
    }
    
    /** Handles the creation of a graphNode. If the node already exists it will
     * return the existing instance. If it does not exist it will create it.
     *
     * Node creation is recursive if there are any transitions coming out of the
     * node.
     */
    protected graphNode createNode(LTSNode n) {
        if (nodeMap.containsKey(n.toString())) {
            // The node already exists, return it
            return nodeMap.get(n.toString());
        }
        else {
            // The node doesnt exist, create it.
            graphNode gn = new graphNode(n);
            gn.randomizePosition();
            System.out.println("   Node create: " + n.toString() + " at position " + gn.position.toString());
            Collection<LTSTransition> ltst = theLTS.getTransitions();
            
            for (LTSTransition l:ltst) {
                if (l.getNode1().equals(n)) {
                    if (l.getNode2() != null) {
                        graphNode nnn = createNode(l.getNode2());
                        nnn.springs.add(gn);
                        nnn.spring_names.add(l.getName());
                        gn.springs.add(nnn);
                        gn.spring_names.add(l.getName());
                        storeNode(nnn);
                    }
                }
            }
            
            return gn;
        }
    }
   
    /** Stores a node for future use.
     */
    protected void storeNode(graphNode n) {
        if (n == null) return;
        theNodes.add(n);
        nodeMap.put(n.theNode.toString(),n);
    }
    
    protected void clearNodes() {
        theNodes.clear();
        nodeMap.clear();
    }
    
    /** Generates position data for the nodes. [[ DEPRICATED ]]
     */
    protected void runSim() {
        System.out.println("Running Simulation...");
         do {
             total_kinetic_energy = 0;
             
             total_kinetic_energy = iterate();
             repaint();

         } while (total_kinetic_energy > 50);
         // Above: Loop until total kinetic energy is less than some small number (the simulation has stopped running)

    }
    
    public double iterate() {
        double total_kinetic_energy = 0;
        graphVector net_force;
        for (graphNode n:theNodes) {
                 net_force = new graphVector(0,0);
                 
                 // For each other node...
                 for (graphNode nn:theNodes) {
                     if (nn != n) {
                         // net_force := net_force + Coulomb_repulsion( this_node, other_node )
                         net_force.add(Coulomb_repulsion(n,nn));
                     }
                 }
                 
                 // For each spring connected to this node...
                 for (graphNode spring:n.springs) {
                     // net_force := net_force + Hooke_attraction( this_node, spring )
                     net_force.add(Hooke_attraction(n,spring)); 
                 }
                 
                 // Without damping, it moves forever:
                 // this_node.velocity := (this_node.velocity + (timestep * net_force)) * damping
                 // timestep * net_force
                 graphVector vma = net_force.scalarMultiply(net_force,timestep);
                 // velocity + (timestep * net_force)
                 graphVector vmb = graphVector.add(n.velocity,vma);
                 // (this_node.velocity + (timestep * net_force)) * damping
                 vmb.scalarMultiply(damping);
                 n.velocity = vmb;
                 /*n.velocity = n.net_force.scalarMultiply(
                         n.net_force.add(n.velocity,n.net_force.scalarMultiply(n.net_force,timestep)),
                         damping);*/
                 
                 // this_node.position := this_node.position + (timestep * this_node.velocity)
                 n.position.add(graphVector.scalarMultiply(n.velocity,timestep));
                 
                 
                 // total_kinetic_energy := total_kinetic_energy + (this_node.velocity)^2
                 total_kinetic_energy += Math.pow(n.velocity.length(),2);
             }
        
        repaint();
        return total_kinetic_energy;
    }
    
    /** Needed by runSim()
     */
    protected graphVector Coulomb_repulsion(graphNode a, graphNode b) {
        graphVector gva = new graphVector (a.position.x - b.position.x, a.position.y - b.position.y);
        double force = repelconst * (200) / Math.pow (gva.length(), 2);
        graphVector gvau = gva.getUnit();
        gvau.scalarMultiply(force);
        return gvau;
    }
    /** Needed by runSim()
     */
    protected graphVector Hooke_attraction(graphNode a, graphNode spring) {
        graphVector d = new graphVector ( a.position.x -spring.position.x,a.position.y - spring.position.y);//   
        d.x *= -springconst;
        d.y *= -springconst;

        return d;
    }
    
    /** Renders the calculated node graph thing
     */
    protected void renderGraph() {
        System.out.println("  Rendering...");
        paintComponent(this.getGraphics());
        
        for (graphNode n:theNodes) {
            System.out.println("  Node '" + n.theNode.toString() + "' is at position '" + n.position.toString() + "'");
        }
        
    }
    
    public void paintComponent(Graphics g) {
        if (theLTS == null) return;
        
        super.paintComponent(g);
        
        g.drawString("Total Kinetic Energy: " + total_kinetic_energy,20,20);
        
        if (theLTS.getNodes().size() == 0) {
            System.out.println("LTSRender: nothing to render.");
            return;
        }
        
        // Draw nodes
        for (graphNode n:theNodes) {
            drawNode(g,n);
        }
        
        // Draw Transitions
        Collection<LTSTransition> transitions = theLTS.getTransitions();
        for (LTSTransition t:transitions) {
            if (t.getNode2() == null) continue;
            drawSpring(g,createNode(t.getNode1()),createNode(t.getNode2()),t.getName());
        }
    }

    protected void drawNode(Graphics g, graphNode gn) {
        
        if (gn.noDraw) return;
        
        Point p = new Point();
        
        p.setLocation((gn.position.x - 5),(gn.position.y - 5));

        //System.out.println("Drawing node " + gn.position.toString() + " at position (" + p.x + "," + p.y + ")");
        
        // draw the node
        g.drawOval(p.x, p.y, 10,10);
        p.setLocation(gn.position.x + 10, gn.position.y + 10);
        g.drawString(gn.theNode.getNodeName(), p.x, p.y);
        
        if (gn.theNode.equals(theLTS.getStartNode())) {
            // Its the start node.
            g.drawOval(p.x-17,p.y - 17,14,14);
        }
        
    }
    
    protected void drawSpring(Graphics g, graphNode g1, graphNode g2, String name) {
        
        if (g1.noDraw || g2.noDraw) return;

        // Draw the arrow
        Point p1 = new Point();
        Point p2 = new Point();
        
        p1.setLocation(g1.position.x,g1.position.y);
        p2.setLocation(g2.position.x,g2.position.y);
        
        Double xx = (g1.position.x + g2.position.x)/2;
        Double yy = (g1.position.y + g2.position.y)/2;
        Point p3 = new Point();
        p3.setLocation(xx,yy);
        
        g.drawString(name,p3.x,p3.y);
        
        drawArrow(g,p1,p2);
        
    }
    
    protected void drawArrow(Graphics g, Point p1, Point p2) {
        // first draw the shaft of the arrow:
        //g.drawLine(p1.x,p1.y,p2.x,p2.y);

        int x0 = p1.x;
        int y0 = p1.y;
        int x1 = p2.x;
        int y1 = p2.y;
        
        int deltaX = x1 - x0;
        int deltaY = y1 - y0;
        double frac = 0.05;

        g.drawLine(x0,y0,x1,y1);
        g.drawLine(x0 + (int)((1-frac)*deltaX + frac*deltaY),  y0 + (int)((1-frac)*deltaY - frac*deltaX),  x1, y1);
        g.drawLine(x0 + (int)((1-frac)*deltaX - frac*deltaY), y0 + (int)((1-frac)*deltaY + frac*deltaX), x1, y1);
    }
}
