/*
 * Main.java
 *
 * Created on 23 July 2007, 21:07
 *
 * To change this template, choose Tools | Template Manager
 * and open the template in the editor.
 */

package assignment;

/* Required packages */
import java.lang.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.JFileChooser;


/**
 *
 * @author David
 */
public class Main implements ActionListener {
    
    protected JFrame window;        // The main window.
    protected LTSView theView;      // Displays an LTS
    protected LTS currentLTS;       // The currently open LTS.
    
    /** Creates a new instance of Main 
     */
    public Main() {
        window = new JFrame("COMP242 Assignment 1");
        window.setSize(800,600);
        window.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        
        Menu theMenu = new Menu(this);

        window.setLayout(new BorderLayout());
        
        window.setJMenuBar(theMenu);
        
        
        // Create a new LTS View panel.
        theView = new ForceLTSView();
        window.add(theView, BorderLayout.CENTER);
        
        JToolBar tb = new JToolBar("Exciting buttons of doom");
        addTBButtons(tb);
        window.add(tb,BorderLayout.PAGE_START);
        
        
        window.setVisible(true);
        
        System.out.println("Autocreating LTS");
        fileNew();
    }
    
    public void addTBButtons(JToolBar tb) {
        JButton button = null;
        
        button = new JButton();
        button.setText("Open");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Compose");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Redraw");
        button.addActionListener(this);
        tb.add(button);
        
        button = new JButton();
        button.setText("Iterate");
        button.addActionListener(this);
        tb.add(button);
    }
    
    public void actionPerformed(ActionEvent event){
        String s = event.getActionCommand();
        
        if (s == "Open")
            handleMenuEvent("Open...");
        if (s == "Compose")
            handleMenuEvent("Compose with...");
        if (s == "Redraw")
            handleMenuEvent("Refresh");
        if (s == "Iterate")
            handleMenuEvent("Sim Iterate");
    }
    
    /** Processes a menu click event.
     */
    public void handleMenuEvent(String menuItemPressed) {
        if (menuItemPressed == "New")
            fileNew();
        else if (menuItemPressed == "Open...")
            fileOpen();
        else if (menuItemPressed == "Save")
            fileSave();
        else if (menuItemPressed == "Save As...")
            fileSaveAs();
        else if (menuItemPressed == "Compose with...")
            fileCompose();
        else if (menuItemPressed == "Exit")
            exit();
        else if (menuItemPressed == "Refresh")
            updateDisplay();
        else if (menuItemPressed == "Sim Iterate")
            iterate();
        else             
            System.out.println("Error: Unknown menu event name \"" + menuItemPressed + "\"");
    }
    
    /** Creates a new LTS.
     */
    private void fileNew() {
        System.out.println("Creating new LTS...");
        currentLTS = new LTS();
        updateTitle("Untitled");
        //updateDisplay();
    }
    
    /** Composes the current LTS with another. First asks the user to select the
     * file containing the LTS to compose with the current one. Then asks the
     * user to enter the Sync Set.
     */
    private void fileCompose() {
        LTS ltsA = currentLTS;
        fileOpen();
        LTS ltsB = currentLTS;
        
        // Ok, now we need to get the Sync set.
        String syncSet = JOptionPane.showInputDialog("Enter Sync Set:");
        
        currentLTS = LTS.composeA(ltsA, ltsB, syncSet);
        currentLTS.setFileName(ltsA.getFileName());
        
        theView.setLTS(currentLTS);
        theView.setup();
    }
    
    /** Opens an LTS from a text file.
     */
    private void fileOpen() {
        JFileChooser jfc = new JFileChooser();
        int returnVal = jfc.showOpenDialog(jfc);
       
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            /* Load the file */
            File file = jfc.getSelectedFile();
            try {
                currentLTS = LTS.load(file.getAbsolutePath());
                updateTitle(file.getAbsolutePath());
                
                theView.setLTS(currentLTS);
                theView.setup();
            }
            catch (IOException e) {
                System.out.println("File load failed: IO Exception");
            }
            
            /* Update the display */
            //updateDisplay();
        }
        else {
            System.out.println("File open operation canceled by user.");
        }
    }
    
    /** Saves the current LTS.
     */
    private void fileSave() {
        try {
        currentLTS.save();
        }
        catch (IOException e) {
            System.out.println("File Save failed: " + e.toString());
        }
    }
    
    /** Saves the current LTS with a different file name.
     */
    private void fileSaveAs() {
        JFileChooser jfc = new JFileChooser();
        int returnVal = jfc.showSaveDialog(jfc);
       
        if (returnVal == JFileChooser.APPROVE_OPTION) {
            File file = jfc.getSelectedFile();
            try {
                currentLTS.save(file.getAbsolutePath());
                updateTitle(file.getAbsolutePath());
            }
            catch (IOException e) {
                System.out.println("File Save failed: " + e.toString());
            }
        } else {
            System.out.println("File save operation canceled by user.");
        }
    }
     
    /** Exits the program.
     */
    private void exit() {
        System.out.println("Application terminated by user.");
        System.exit(0);     /* Terminate Normally */
    }
    
    /** Updates the display (recalculates/redraws the LTS)
     */
    private void updateDisplay() {
        System.out.println("Updating display...");
        theView.renderLTS();
        System.out.println("Display updated.");
    }
    
    private void updateTitle(String addition) {
        window.setTitle("COMP242 Assignment 1 - [" + addition + "]");
    }
    
    private void iterate() {
        Double val = theView.iterate();
        updateTitle(val.toString());
    }
    
    /** Program entry point.
     * @param args the command line arguments
     */
    public static void main(String[] args) {
        System.out.println("COMP242 Assignment 1\n     (C) Copyright David Goodwin, 2007");
        System.out.println("\nProgram Starting...");
        Main theApp = new Main();
    }
    
}
